import warnings

from django.contrib import admin

from .forms import TranslationFormSet
from . import settings


class TranslationInlineMixin(object):
    def __init__(self, *args, **kwargs):
        super(TranslationInlineMixin, self).__init__(*args, **kwargs)

        if settings.HIDE_LANGUAGE:
            self.exclude = ('language_code', )
            self.can_delete = False

    extra = 1
    formset = TranslationFormSet
    max_num = len(settings.LANGUAGES)


class TranslationStackedInline(TranslationInlineMixin, admin.StackedInline):
    pass


class TranslationTabularInline(TranslationInlineMixin, admin.TabularInline):
    pass


class TranslationInline(TranslationStackedInline):
    def __init__(self, *args, **kwargs):
        warnings.warn(DeprecationWarning(
            "TranslationInline is deprecated; "
            "use TranslationStackedInline or TranslationTabularInline instead."
        ))

        return super(TranslationInline, self).__init__(*args, **kwargs)



import logging

logger = logging.getLogger(__name__)

from django.utils.translation import ugettext_lazy as _
from django.forms.models import BaseInlineFormSet
from django import forms

from . import settings


class TranslationFormSet(BaseInlineFormSet):
    """
    FormSet for TranslationInlines, making sure that at least one translation
    is required and that sensible default values are selected for the language
    choice.
    """

    def clean(self):
        """
        Make sure there is at least a translation has been filled in. If a
        default language has been specified, make sure that it exists amongst
        translations.
        """

        # First make sure the super's clean method is called upon.
        super(TranslationFormSet, self).clean()

        if settings.HIDE_LANGUAGE:
            return

        if len(self.forms) > 0:
            # If a default language has been provided, make sure a translation
            # is available

            if settings.DEFAULT_LANGUAGE and not any(self.errors):
                # Don't bother validating the formset unless each form is
                # valid on its own. Reference:
                # http://docs.djangoproject.com/en/dev/topics/forms/formsets/#custom-formset-validation

                for form in self.forms:
                    language_code = form.cleaned_data.get(
                        'language_code', None
                    )

                    if language_code == settings.DEFAULT_LANGUAGE:

                        # All is good, don't bother checking any further
                        return

                raise forms.ValidationError(_(
                    'No translation provided for default language \'%s\'.'
                ) % settings.DEFAULT_LANGUAGE)

        else:
            raise forms.ValidationError(
                _('At least one translation should be provided.')
            )

    def _construct_available_languages(self):
        self.available_languages = [
            choice[0]
            for choice in self.form.base_fields['language_code'].choices
            if choice[0] != ''
        ]

    def _construct_forms(self):
        """
        Before we're constructing forms, make sure a complete list of
        languages is available. This is used to select sensible defaults
        for the language_code field.
        """
        if not settings.HIDE_LANGUAGE:
            self._construct_available_languages()

        super(TranslationFormSet, self)._construct_forms()

    def _get_default_language(self):
        """
        If a default language has been set, and is still available in
        `self.available_languages`, return it and remove it from the list.

        If not, simply pop the first available language.
        """

        assert hasattr(self, 'available_languages'), \
            'No available languages have been generated.'
        assert len(self.available_languages) > 0, \
            'No available languages to select from.'

        if (
            settings.DEFAULT_LANGUAGE and
            settings.DEFAULT_LANGUAGE in self.available_languages
        ) or (
            'language_code' not in self.form.base_fields
        ):
            # Default language still available

            self.available_languages.remove(settings.DEFAULT_LANGUAGE)
            return settings.DEFAULT_LANGUAGE

        else:
            # Select the first item and return it
            return self.available_languages.pop(0)

    def _construct_form(self, i, **kwargs):
        """
        Construct the form, overriding the initial value for `language_code`.
        """
        form = super(TranslationFormSet, self)._construct_form(i, **kwargs)

        if settings.HIDE_LANGUAGE:
            form.instance.language_code = settings.DEFAULT_LANGUAGE
        else:
            language_code = form.instance.language_code

            if language_code:
                logger.debug(
                    u'Removing translation choice %s for instance %s'
                    u' in form %d', language_code, form.instance, i
                )

                self.available_languages.remove(language_code)

            else:
                initial_language_code = self._get_default_language()

                logger.debug(
                    u'Preselecting language code %s for form %d',
                    initial_language_code, i
                )

                form.initial['language_code'] = initial_language_code

        return form



import logging
logger = logging.getLogger('multilingual_model')

import re

from django.utils.translation import ugettext_lazy as _
from django.utils.translation import ugettext

from django.db import models
from django.utils.translation import get_language
from django.core.exceptions import ObjectDoesNotExist

from . import settings

# Match something like en, but also en_us
LANGUAGE_CODE_RE = re.compile(
    r'_(?P<base_code>[a-z]{2,7})(_(?P<ext_code>[a-z]{2,7})){0,1}$'
)


class MultilingualTranslation(models.Model):
    """ Abstract base class for translations. """

    class Meta:
        abstract = True

    language_code = models.CharField(
        _('language'), max_length=7, choices=settings.LANGUAGES,
        blank=False, null=False
    )


class MultilingualModel(models.Model):
    """ Provides support for multilingual fields. """

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        super(MultilingualModel, self).__init__(*args, **kwargs)
        self._language = get_language()
        self._translation_cache = {}

    def _get_translation(self, field, code):
        """
        Gets the translation of a specific field for a specific language code.

        This raises ObjectDoesNotExist if the lookup was unsuccesful. As of
        today, this stuff is cached. As the cache is rather aggressive it
        might cause rather strange effects. However, we would see the same
        effects when an ordinary object is changed which is already in memory:
        the old state would remain.
        """

        if not code in self._translation_cache:
            translations = self.translations.select_related()

            logger.debug(
                u'Matched with field %s for language %s. Attempting lookup.',
                field, code
            )

            try:
                translation_obj = translations.get(language_code=code)

            except ObjectDoesNotExist:
                translation_obj = None

            self._translation_cache[code] = translation_obj

            logger.debug(u'Translation not found in cache.')

        else:
            logger.debug(u'Translation found in cache.')
            # Get the translation from the cache
            translation_obj = self._translation_cache.get(code)

        # If this is none, it means that a translation does not exist
        # It is important to cache this one as well
        if not translation_obj:
            raise ObjectDoesNotExist

        field_value = getattr(translation_obj, field)

        logger.debug(
            u'Found translation object %s, returning value %s.',
            translation_obj, field_value
        )

        return field_value

    def __getattr__(self, attr):
        # If an attribute is defined in the current model, don't
        # look any further.
        if attr in self.__dict__:
            # TODO: Test coverage!
            return self.__dict__[attr]

        # See whether we can find a translation for the field
        translated_fields = self.translations.model._meta.get_all_field_names()
        for field in translated_fields:
            code = None

            # Only consider attributes starting with this field name
            if attr.startswith(field):
                # If we have a match, see if we can re-match the language
                # code in the remaining string.
                match = LANGUAGE_CODE_RE.match(attr[len(field):])
                if match:
                    base_code = match.group('base_code')
                    ext_code = match.group('ext_code')

                    # TODO: CLEANUP
                    # This is ugly code and redundant.

                    if ext_code:
                        code = '%s-%s' % (base_code, ext_code)
                    else:
                        code = base_code

                    logger.debug(
                        u'Regular expression match, resulting code: %s', code
                    )

                elif attr in translated_fields:
                    # TODO: Test coverage!

                    code = self._language
                    base_code = None
                    field = attr

                    logger.debug(
                        u'Regular expression not matched but translated'
                        u' field detected.'
                    )

            if code:
                try:
                    return self._get_translation(field, code)

                except ObjectDoesNotExist:
                    # TODO: CLEANUP
                    # This is ugly code and redundant.
                    if not base_code:
                        # TODO: Test coverage!
                        base_pos = code.find('-')
                        if base_pos > 0:
                            base_code = code[:base_pos]

                    if base_code:

                        logger.debug(
                            u'Attempting a match for the base \'%s\'',
                            base_code
                        )

                        try:
                            return self._get_translation(field, base_code)
                        except ObjectDoesNotExist:
                            pass

                    logger.debug(
                        u'Lookup failed, attempting fallback or '
                        u'failing silently.'
                    )

                    # If we're using a default language and the current
                    # language is not the default language (which has already
                    # been checked), lookup the value for the default language.
                    if (
                        settings.FALL_BACK_TO_DEFAULT and
                        settings.DEFAULT_LANGUAGE and
                        code != settings.DEFAULT_LANGUAGE
                    ):

                        try:
                            return self._get_translation(
                                field, settings.DEFAULT_LANGUAGE
                            )

                        except ObjectDoesNotExist:
                            # TODO: Test coverage!
                            pass

                    # TODO: Test coverage!
                    if settings.FAIL_SILENTLY:
                        return None

                    raise ValueError(
                        u"'%s' object with pk '%s' has no"
                        u" translation to '%s'" % (
                            self._meta.object_name, self.pk, code
                        )
                    )

        raise AttributeError(
            u"'%s' object has no attribute '%s'" % (
                self._meta.object_name, str(attr)
            )
        )

    def unicode_wrapper(self, property, default=ugettext('Untitled')):
        """
        Wrapper to allow for easy unicode representation of an object by
        the specified property. If this wrapper is not able to find the
        right translation of the specified property, it will return the
        default value instead.

        Example::
            def __unicode__(self):
                return unicode_wrapper('name', default='Unnamed')

        """
        # TODO: Test coverage!
        try:
            value = getattr(self, property)
        except ValueError:
            logger.warn(
                u'ValueError rendering unicode for %s object.',
                self._meta.object_name
            )

            value = None

        if not value:
            value = default

        return value



from django.conf import settings


FAIL_SILENTLY = getattr(
    settings, 'MULTILINGUAL_FAIL_SILENTLY', not settings.DEBUG
)

DEFAULT_LANGUAGE = getattr(
    settings, 'MULTILINGUAL_LANGUAGE_CODE', settings.LANGUAGE_CODE
)

FALL_BACK_TO_DEFAULT = getattr(
    settings, 'MULTILINGUAL_FALL_BACK_TO_DEFAULT', True
)

LANGUAGES = getattr(
    settings, 'MULTILINGUAL_LANGUAGES', settings.LANGUAGES
)

HIDE_LANGUAGE = getattr(
    settings, 'MULTILINGUAL_HIDE_LANGUAGE', len(LANGUAGES) == 1
)



from django.db import models
from django.test import TestCase
from django.utils import translation

from .models import MultilingualModel, MultilingualTranslation


class BookTranslation(MultilingualTranslation):
    parent = models.ForeignKey('Book', related_name='translations')

    title = models.CharField(max_length=32)
    description = models.TextField()


class Book(MultilingualModel):
    ISBN = models.IntegerField()


__test__ = {'doctest': """
>>> book = Book(ISBN="1234567890")
>>> book.save()
>>> book_en = BookTranslation(language_code='en')
>>> book_en.title = "Django for Dummies"
>>> book_en.description = "Django described in simple words."
>>> book_en.parent = book
>>> book_en.save()
>>> book_pl = BookTranslation(language_code='pl')
>>> book_pl.title = "Django dla Idiotow"
>>> book_pl.description = "Django opisane w prostych slowach"
>>> book_pl.parent = book
>>> book_pl.save()
>>> # now here comes the magic
>>> book.title_en
u'Django for Dummies'
>>> book.description_pl
u'Django opisane w prostych slowach'
>>> mybook = Book(ISBN=1234)
>>> mybook.save()
>>> trans_en = BookTranslation(parent=mybook, language_code='en', \
    title='Django for dummies', description='Django for absolute nitwits.')
>>> trans_en.save()
>>> trans_nl = BookTranslation(parent=mybook, language_code='nl', \
    title='Django voor dummies', description='Django voor complete idioten.')
>>> trans_nl.save()
>>> mybook.title_nl
u'Django voor dummies'
>>> mybook.title_en
u'Django for dummies'
>>> mybook.description_en
u'Django for absolute nitwits.'
>>> mybook.description_nl
u'Django voor complete idioten.'
"""}


class BookTestCase(TestCase):
    def setUp(self):
        """ Setup a book with translations. """

        self.book = Book(ISBN="1234567890")
        self.book.save()

        self.book_en = BookTranslation(language_code='en')
        self.book_en.title = "Django for Dummies"
        self.book_en.description = "Django described in simple words."
        self.book_en.parent = self.book
        self.book_en.save()

        self.book_en_us = BookTranslation(language_code='en-us')
        self.book_en_us.title = "Django for Dumbo's"
        self.book_en_us.description = "Django described in simple woords."
        self.book_en_us.parent = self.book
        self.book_en_us.save()

        self.book_pl = BookTranslation(language_code='pl')
        self.book_pl.title = "Django dla Idiotow"
        self.book_pl.description = "Django opisane w prostych slowach"
        self.book_pl.parent = self.book
        self.book_pl.save()

    def test_explicit(self):
        """
        Test explicit request for a specific language. This is basically
        a mockup of the original docstring test.
        """

        self.assertEquals(self.book.title_en, self.book_en.title)
        self.assertEquals(self.book.title_pl, self.book_pl.title)

    def test_language_detection(self):
        """
        Set the language to 'pl' and test whether this is detected from
        within the model.
        """

        for test_lang in ('pl', 'en'):
            # Set the language
            translation.activate(test_lang)

            # Get the book
            book = Book.objects.get(ISBN=self.book.ISBN)

            # Check if the language set in book's init is actually the right
            # language.
            self.assertEquals(book._language, test_lang)

    def test_base_locale_default(self):
        """
        Explicitly request the title in a specific sublocale (en-us) and
        check whether we will get the available base locale instead.
        """

        # assert equality of self.book_en.title and self.book.title_en_us
        self.assertEquals(self.book_en.title, self.book.title_en_kk)

    def test_base_locale_explicit(self):
        """
        Test whether we can explicitly get the value of a string in
        a specific sub-locale - through something like book.title_en_us.
        """

        # Assert equality of book.title_en_us to this unique value
        self.assertEqual(self.book.title_en_us, self.book_en_us.title)

    def test_nonexisting_field(self):
        """
        Check whether requesting an unknown field actually raises an
        AttributeError.
        """

        # assert an exception for self.book.bananas
        with self.assertRaises(AttributeError):
            self.book.bananas

    def test_nonexisting_translation(self):
        """
        Test whether requesting a non-existing translation (when no
        default translation has been used) raises an AttributeError. If
        `FALL_BACK_TO_DEFAULT` and `DEFAULT_LANGUAGE` have been specified -
        chekc if we're actually returning the default.
        """

        # Only do this if no default translation has been defined
        from multilingual_model import settings

        if settings.FALL_BACK_TO_DEFAULT and settings.DEFAULT_LANGUAGE:
            # See if we're given the default language

            result = self.book.title_dk

            attribute = \
                'title_%s' % settings.DEFAULT_LANGUAGE.replace('-', '_')

            self.assertEqual(getattr(self.book, attribute), result)

        else:
            # No defaults have been specifed

            # assert an exception for self.book.title_kk
            with self.assertRaises(AttributeError):
                self.book.title_dk

    def test_serbian_latin(self):
        """ Test for language code edge case. """

        test_lang = 'sr-latn'

        book_translation = BookTranslation(language_code=test_lang)
        book_translation.title = "Django dla Idiotow"
        book_translation.description = "Django opisane w prostych slowach"
        book_translation.parent = self.book
        book_translation.save()

        # Set the language
        translation.activate(test_lang)

        # Get the book
        book = Book.objects.get(ISBN=self.book.ISBN)

        # Check if the language set in book's init is actually the right
        # language.
        self.assertEquals(book._language, test_lang)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#from distribute_setup import use_setuptools
#use_setuptools()

from setuptools import setup, find_packages

try:
    README = open('README.rst').read()
except:
    README = None

#try:
#    REQUIREMENTS = open('requirements.txt').lines()
#except:
#    REQUIREMENTS = None

setup(
    name='django-multilingual-model',
    version='0.5.1',
    description='Django Simple Multilingual Support for Models.',
    long_description=README,
    #install_requires=REQUIREMENTS,
    author='Mathijs de Bruin',
    author_email='mathijs@mathijsfietst.nl',
    url='http://github.com/dokterbob/django-multilingual-model',
    packages = find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    test_suite='setuptest.setuptest.SetupTestSuite',
    tests_require=(
        'django-setuptest',
        'argparse',  # apparently needed by django-setuptest on python 2.6
    ),
)
